/*
QBL.cpp
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "QBL.h"

namespace qbl
{
void ActionType::add(wok::interp::Object& A, wok::interp::Object& B, wok::interp::Object& output)
{
	wok::interp::UserObject *o1, *o2;
	o1 = wok::interp::ObjectHelper::getUserObject(A);
	o2 = wok::interp::ObjectHelper::getUserObject(B);

	if(o1->getType() == "qbl::qdm::Action" && o2->getType() == "qbl::qdm::Action")
	{
		// action + action = ActionList
		ActionListType *t = new ActionListType;
		t->actions.push_back(((ActionType*)o1)->a);
		t->actions.push_back(((ActionType*)o2)->a);

		wok::interp::ObjectHelper::newUserObject(t, output);
	}
	else if(o1->getType() == "qbl::qdm::ActionList" && o2->getType() == "qbl::qdm::Action")
	{
		ActionListType *t = new ActionListType;
		t->actions = ((ActionListType*)o1)->actions;
		t->actions.push_back(((ActionType*)o2)->a);

		wok::interp::ObjectHelper::newUserObject(t, output);
	}
}

void ActionListType::add(wok::interp::Object& A, wok::interp::Object& B, wok::interp::Object& output)
{
	wok::interp::UserObject *o1, *o2;
	o1 = wok::interp::ObjectHelper::getUserObject(A);
	o2 = wok::interp::ObjectHelper::getUserObject(B);

	if(o1->getType() == "qbl::qdm::ActionList" && o2->getType() == "qbl::qdm::ActionList")
	{
		ActionListType *t = new ActionListType;
		t->actions = ((ActionListType*)o1)->actions;
		for(int i=0; i<(int)((ActionListType*)o2)->actions.size(); ++i)
			t->actions.push_back(((ActionListType*)o2)->actions[i]);

		wok::interp::ObjectHelper::newUserObject(t, output);
	}
	else if(o1->getType() == "qbl::qdm::ActionList" && o2->getType() == "qbl::qdm::Action")
	{
		ActionListType *t = new ActionListType;
		t->actions = ((ActionListType*)o1)->actions;
		t->actions.push_back(((ActionType*)o2)->a);

		wok::interp::ObjectHelper::newUserObject(t, output);
	}
}

bool QBLCommands::evaluate(wok::Text& text, wok::interp::Store& store, wok::interp::Object& output, wok::interp::Interpreter* interpreter, std::vector <wok::interp::CommandLibrary*> &library, bool globalMode)
{
	if(text[0] == "qbl::dist")
	{
		// construct a new distribution type of the form { (double pr) (integer actionSpaceIndex) (integer action) }^* (if arg = 0, then blank dist type)
		if(text.size() == 2)
		{
			wok::interp::ObjectHelper::newUserObject(new DistributionType, output);
		}
		//else if((text.size()-1) % 3 == 0)
		else
		{
			DistributionType *o = new DistributionType;
			for(int i=1; i<(int)text.size(); i+=3)
			{
				if(i+2 >= (int)text.size()) break;
				wok::interp::Object pr_obj, actionSpace_obj, action_obj;
				interpreter->evaluate(text[i], store, pr_obj, library, globalMode);
				interpreter->evaluate(text[i+1], store, actionSpace_obj, library, globalMode);
				interpreter->evaluate(text[i+2], store, action_obj, library, globalMode);

				qdm::TrainingDistribution p;
				if(pr_obj.getType() == wok::interp::Double)
					p.p = *wok::interp::ObjectHelper::getDouble(pr_obj);
				else
					p.p = *wok::interp::ObjectHelper::getInteger(pr_obj);
				qdm::ActionData data;
				data.actionSpaceIndex = *wok::interp::ObjectHelper::getInteger(actionSpace_obj);
				data.actionIndex = *wok::interp::ObjectHelper::getInteger(action_obj);
				p.actionList.push_back(data);
				o->p.push_back(p);
			}

			wok::interp::ObjectHelper::newUserObject(o, output);
		}
		/*else
		{
			std::cout << "qbl::dist argument error.\n";
			return false;
		}*/

		return true;
	}
	else if(text[0] == "qbl::dist_str")
	{
		// construct a new distribution type of the form { (double pr) {(string action)}^+ }^* (if arg = 0, then blank dist type)
		if(text.size() == 2)
		{
			wok::interp::ObjectHelper::newUserObject(new DistributionType, output);
		}
		else
		{
			wok::interp::Object qdm_obj;
			interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
			if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
			{
				std::cout << "qbl::dist_str syntax error (first argument must be of type qbl::QDM).\n";
				return false;
			}

			qdm::QDM* qdm = ((QDMType*)qdm_obj.getValue())->qdm;
			if(!qdm)
				return false;

			DistributionType *o = new DistributionType;
			//for(int i=2; i<(int)text.size(); i+=2)
			int i = 2;
			while(i < (int)text.size())
			{
				if(i+1 >= (int)text.size()) break;
				wok::interp::Object pr_obj, action_obj;
				interpreter->evaluate(text[i], store, pr_obj, library, globalMode);
				++i;
				qdm::TrainingDistribution p;
				if(pr_obj.getType() == wok::interp::Double)
					p.p = *wok::interp::ObjectHelper::getDouble(pr_obj);
				else
					p.p = *wok::interp::ObjectHelper::getInteger(pr_obj);

				while(i < (int)text.size())
				{
					qdm::ActionData data;
					interpreter->evaluate(text[i], store, action_obj, library, globalMode);
					if(action_obj.getType() == wok::interp::Double) break;
					if(wok::interp::ObjectHelper::getString(action_obj)->size() == 0) break;

					if(!qdm->lookupAction(*wok::interp::ObjectHelper::getString(action_obj), data.actionSpaceIndex, data.actionIndex))
					{
						std::cout << "Error: action '" << *wok::interp::ObjectHelper::getString(action_obj) << "' not defined.\n";
						break;
					}

					p.actionList.push_back(data);
					++i;
				}
				//p.actionSpace = *wok::interp::ObjectHelper::getInteger(action_obj);
				//p.action = *wok::interp::ObjectHelper::getInteger(action_obj);
				o->p.push_back(p);
			}

			wok::interp::ObjectHelper::newUserObject(o, output);
		}

		return true;
	}
	else if(text[0] == "qbl::rand")
	{
		wok::interp::Object max_obj;
		interpreter->evaluate(text[1], store, max_obj, library, globalMode);
		wok::interp::ObjectHelper::newInteger(rand()%(*wok::interp::ObjectHelper::getInteger(max_obj)), output);
	}
	else if(text[0] == "qbl::SensorList")
	{
		SensorListType *o = new SensorListType;
		for(int i=1; i<(int)text.size(); ++i)
		{
			wok::interp::Object str;
			interpreter->evaluate(text[i], store, str, library, globalMode);

			o->v.push_back(*wok::interp::ObjectHelper::getString(str));
		}

		wok::interp::ObjectHelper::newUserObject(o, output);
		return true;
	}
	else if(text[0] == "qbl::State")
	{
		if(text.size() < 2)
		{
			std::cout << "Error: Arguments for qbl::State are MODE:STRING {qbl::SensorList}^*\n";
			return false;
		}
		StateType *o = new StateType;

		wok::interp::Object mode_obj;
		interpreter->evaluate(text[1], store, mode_obj, library, globalMode);

		if(mode_obj.getType() == wok::interp::String)
		{
			if(*wok::interp::ObjectHelper::getString(mode_obj) == "repeat")
				o->state.mode = qdm::TrainingSensorSet::Repeat;
			else if(*wok::interp::ObjectHelper::getString(mode_obj) == "single")
				o->state.mode = qdm::TrainingSensorSet::SingleUse;
			else
			{
				delete o;
				return false;
			}
		}
		else
		{
			delete o;
			return false;
		}

		for(int i=2; i<(int)text.size(); ++i)
		{
			wok::interp::Object sensor_obj;
			interpreter->evaluate(text[i], store, sensor_obj, library, globalMode);
			if(wok::interp::ObjectHelper::getUserObject(sensor_obj)->getType() == "qbl::SensorList")
			{
				SensorListType *t = (SensorListType*)wok::interp::ObjectHelper::getUserObject(sensor_obj);
				o->state.sensors.push_back(t->v);
			}
			else
			{
				delete o;
				return false;
			}
		}

		wok::interp::ObjectHelper::newUserObject(o, output);
		return true;
	}
	else if(text[0] == "qbl::add_rule")
	{
		// syntax is "qbl::add_rule (TRAINING_SET) (STATE_TYPE) (DIST)
		wok::interp::Object set_obj, state_obj, dist_obj;
		interpreter->evaluate(text[1], store, set_obj, library, globalMode);
		interpreter->evaluate(text[2], store, state_obj, library, globalMode);
		interpreter->evaluate(text[3], store, dist_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(set_obj)->getType() != "qbl::TrainingSet" ||
			wok::interp::ObjectHelper::getUserObject(state_obj)->getType() != "qbl::State" ||
			wok::interp::ObjectHelper::getUserObject(dist_obj)->getType() != "qbl::Distribution")
		{
			std::cout << "qbl::add_rule syntax error.\n";
			return false;
		}

		TrainingSetType *t = new TrainingSetType;
		t->set = ((TrainingSetType*)set_obj.getValue())->set;

		qdm::TrainingSet set;
		set.states = ((StateType*)state_obj.getValue())->state;
		set.distribution = ((DistributionType*)dist_obj.getValue())->p;

		t->set.push_back(set);

		wok::interp::ObjectHelper::newUserObject(t, output);
		return true;
	}
	else if(text[0] == "qbl::train")
	{
		// syntax is "qbl::train (QDM) (TRAINING_SET) (NUM_ITERATIONS) (MEASURE_TIME) (START_POPULATION) (MAX_POPULATION)
		wok::interp::Object qdm_obj, set_obj, iter_obj, measureTime_obj, startPopulation_obj, maxPopulation_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, set_obj, library, globalMode);
		interpreter->evaluate(text[3], store, iter_obj, library, globalMode);
		interpreter->evaluate(text[4], store, measureTime_obj, library, globalMode);
		interpreter->evaluate(text[5], store, startPopulation_obj, library, globalMode);
		interpreter->evaluate(text[6], store, maxPopulation_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM" ||
			wok::interp::ObjectHelper::getUserObject(set_obj)->getType() != "qbl::TrainingSet")
		{
			std::cout << "qbl::train syntax error.\n";
			return false;
		}

		qdm::QDM* qdm = ((QDMType*)qdm_obj.getValue())->qdm;
		if(!qdm)
			return false;

		qdm->setupTraining(*wok::interp::ObjectHelper::getInteger(startPopulation_obj), *wok::interp::ObjectHelper::getInteger(maxPopulation_obj));
		qdm->prepare();

		int numIterations = *wok::interp::ObjectHelper::getInteger(iter_obj);
		int measureTime = *wok::interp::ObjectHelper::getInteger(measureTime_obj);
		for(int i=0; i<numIterations; ++i)
		{
#ifdef WIN32
			if(kbhit() && getch() == 'q') break;
#endif
			qdm->train(((TrainingSetType*)set_obj.getValue())->set, measureTime, i==0);
		}

		return true;
	}
	else if(text[0] == "qbl::add_sensor")
	{
		// syntax is "qbl::add_sensor (QDM) (SENSOR_NAME)"
		wok::interp::Object qdm_obj, name_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, name_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::add_sensor syntax error: first agument must be of type 'qbl::QDM'\n";
			return false;
		}

		qdm::QDM* qdm = ((QDMType*)qdm_obj.getValue())->qdm;
		if(qdm)
		{
			qdm::Sensor sensor;
			sensor.name = std::string(*wok::interp::ObjectHelper::getString(name_obj));
			qdm->addSensor(sensor);
		}
		return true;
	}
	else if(text[0] == "qbl::save")
	{
		wok::interp::Object qdm_obj, filename_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, filename_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::save syntax error: first argument must be of type 'qbl::QDM'\n";
			return false;
		}

		qdm::QDM* qdm = ((QDMType*)qdm_obj.getValue())->qdm;
		if(qdm)
		{
			std::cout << "Saving sensors to file '" << *wok::interp::ObjectHelper::getString(filename_obj) << "'\n";
			qdm->saveToFile(*wok::interp::ObjectHelper::getString(filename_obj));
		}
		return true;
	}
	else if(text[0] == "qbl::load")
	{
		wok::interp::Object qdm_obj, filename_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, filename_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::save syntax error: first argument must be of type 'qbl::QDM'\n";
			return false;
		}

		qdm::QDM* qdm = ((QDMType*)qdm_obj.getValue())->qdm;
		if(qdm)
		{
			std::cout << "Loading sensors from file '" << *wok::interp::ObjectHelper::getString(filename_obj) << "'\n";
			qdm->loadFromFile(*wok::interp::ObjectHelper::getString(filename_obj));
		}
		return true;
	}
	else if(text[0] == "qbl::qdm::new")
	{
		wok::interp::Object qdm_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::new syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		((QDMType*)qdm_obj.getValue())->qdm = new qdm::QDM;
#ifdef USE_VISUAL_DEBUG
		((QDMType*)qdm_obj.getValue())->qdm->setup(graphClient);
#endif
	}
	else if(text[0] == "qbl::qdm::free")
	{
		wok::interp::Object qdm_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::free syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(((QDMType*)qdm_obj.getValue())->qdm)
			delete ((QDMType*)qdm_obj.getValue())->qdm;
	}
	else if(text[0] == "qbl::qdm::apply_op")
	{
		wok::interp::Object qdm_obj, sensor_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, sensor_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::apply_op syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(((QDMType*)qdm_obj.getValue())->qdm)
			((QDMType*)qdm_obj.getValue())->qdm->applySensor(*wok::interp::ObjectHelper::getString(sensor_obj));
	}
	else if(text[0] == "qbl::qdm::think")
	{
		wok::interp::Object qdm_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::apply_op syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(((QDMType*)qdm_obj.getValue())->qdm)
			((QDMType*)qdm_obj.getValue())->qdm->think();
	}
	else if(text[0] == "qbl::qdm::prepare")
	{
		wok::interp::Object qdm_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::apply_op syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(((QDMType*)qdm_obj.getValue())->qdm)
			((QDMType*)qdm_obj.getValue())->qdm->prepare();
	}
	else if(text[0] == "qbl::qdm::get_max_measure_time")
	{
		wok::interp::Object qdm_obj, index_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, index_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::get_max_measure_time syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(((QDMType*)qdm_obj.getValue())->qdm)
			wok::interp::ObjectHelper::newInteger(((QDMType*)qdm_obj.getValue())->qdm->getMaxMeasureTime(*wok::interp::ObjectHelper::getInteger(index_obj)), output);
	}
	else if(text[0] == "qbl::qdm::decide")
	{
		wok::interp::Object qdm_obj, index_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, index_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::decide syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(((QDMType*)qdm_obj.getValue())->qdm)
		{
			qdm::Point p;
			((QDMType*)qdm_obj.getValue())->qdm->partialMeasure(p, *wok::interp::ObjectHelper::getInteger(index_obj));

			//std::cout << "p = " << p.coord[0] << "\n";
			wok::interp::ObjectHelper::newInteger(((QDMType*)qdm_obj.getValue())->qdm->getAction(p, *wok::interp::ObjectHelper::getInteger(index_obj)), output);
		}
	}
	else if(text[0] == "qbl::qdm::decide_str")
	{
		wok::interp::Object qdm_obj, index_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, index_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::decide syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(((QDMType*)qdm_obj.getValue())->qdm)
		{
			qdm::Point p;
			((QDMType*)qdm_obj.getValue())->qdm->partialMeasure(p, *wok::interp::ObjectHelper::getInteger(index_obj));

			//std::cout << "p = " << p.coord[0] << "\n";
			qdm::Action a;
			((QDMType*)qdm_obj.getValue())->qdm->getAction(p, *wok::interp::ObjectHelper::getInteger(index_obj), &a);
			wok::interp::ObjectHelper::newString(a.name, output);
		}
	}
	else if(text[0] == "qbl::qdm::finalize")
	{
		wok::interp::Object qdm_obj, memory_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, memory_obj, library, globalMode);

		int memorySize = 1;
		if(memory_obj.getType() == wok::interp::Integer)
			memorySize = *wok::interp::ObjectHelper::getInteger(memory_obj);
		if(memorySize <= 0)
		{
			std::cout << "Warning: memorySize <= 0; setting memorySize = 1\n";
			memorySize = 1;
		}

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::finalize syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(((QDMType*)qdm_obj.getValue())->qdm)
		{
			if(((QDMType*)qdm_obj.getValue())->qdm->finalizeSpace(memorySize))
				wok::interp::ObjectHelper::newInteger(1, output);
			else
				wok::interp::ObjectHelper::newInteger(0, output);
		}
	}
	else if(text[0] == "qbl::qdm::add_action_space")
	{
		wok::interp::Object qdm_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::add_action_space syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(!((QDMType*)qdm_obj.getValue())->qdm)
		{
			std::cout << "Error: qbl::QDM = NULL; make sure to qbl::qdm::new first.\n";
			return true;
		}

		for(int i=2; i<(int)text.size(); ++i)
		{
			wok::interp::Object action_obj;
			interpreter->evaluate(text[i], store, action_obj, library, globalMode);
			if(action_obj.getType() == wok::interp::String && i == (int)text.size()-1)
				break;
			if(wok::interp::ObjectHelper::getUserObject(action_obj)->getType() != "qbl::qdm::ActionSpace")
			{
				std::cout << "qbl::qdm::add_action_space syntax error: argument must be of type 'qbl::qdm::ActionSpace'.\n";
				return true;
			}

			((QDMType*)qdm_obj.getValue())->qdm->addActionSpace(((ActionSpaceType*)action_obj.getValue())->A);
		}
	}
	else if(text[0] == "qbl::qdm::ActionList")
	{
		ActionListType *t = new ActionListType;
		for(int i=1; i<(int)text.size(); ++i)
		{
			wok::interp::Object action_obj;
			interpreter->evaluate(text[i], store, action_obj, library, globalMode);
			if(action_obj.getType() == wok::interp::String && i == (int)text.size()-1)
				break;
			if(wok::interp::ObjectHelper::getUserObject(action_obj)->getType() != "qbl::qdm::Action")
			{
				std::cout << "qbl::qdm::ActionList syntax error: argument must be of type 'qbl::qdm::Action'.\n";
				return true;
			}

			t->actions.push_back(((ActionType*)action_obj.getValue())->a);
		}
		wok::interp::ObjectHelper::newUserObject(t, output);
	}
	else if(text[0] == "qbl::qdm::Point")
	{
		PointType *t = new PointType;
		for(int i=1; i<(int)text.size(); ++i)
		{
			wok::interp::Object coord_obj;
			interpreter->evaluate(text[i], store, coord_obj, library, globalMode);
			if(coord_obj.getType() == wok::interp::String && i == (int)text.size()-1)
				break;
			t->p.coord.push_back(*wok::interp::ObjectHelper::getInteger(coord_obj));
		}

		wok::interp::ObjectHelper::newUserObject(t, output);
	}
	else if(text[0] == "qbl::qdm::Action")
	{
		// syntax is Action DIM P1:INT P2:INT P3:INT ...
		// OR Action Point Name
		wok::interp::Object dim_obj, coord_obj;
		interpreter->evaluate(text[1], store, dim_obj, library, globalMode);
		ActionType *t = new ActionType;
		if(dim_obj.getType() == wok::interp::UserType && wok::interp::ObjectHelper::getUserObject(dim_obj)->getType() == "qbl::qdm::Point")
		{
			t->a.p = ((PointType*)dim_obj.getValue())->p;
			wok::interp::Object name_obj;
			interpreter->evaluate(text[2], store, name_obj, library, globalMode);
			t->a.name = *wok::interp::ObjectHelper::getString(name_obj);
		}
		else
		{
			for(int i=0; i<*wok::interp::ObjectHelper::getInteger(dim_obj); ++i)
			{
				interpreter->evaluate(text[i+2], store, coord_obj, library, globalMode);
				t->a.p.coord.push_back(*wok::interp::ObjectHelper::getInteger(coord_obj));
			}
		}

		wok::interp::ObjectHelper::newUserObject(t, output);
	}
	else if(text[0] == "qbl::qdm::shell")
	{
		wok::interp::Object qdm_obj, command_obj;
		interpreter->evaluate(text[1], store, qdm_obj, library, globalMode);
		interpreter->evaluate(text[2], store, command_obj, library, globalMode);

		if(wok::interp::ObjectHelper::getUserObject(qdm_obj)->getType() != "qbl::QDM")
		{
			std::cout << "qbl::qdm::add_action_space syntax error: first argument must be of type 'qbl::QDM'.\n";
			return true;
		}

		if(!((QDMType*)qdm_obj.getValue())->qdm)
		{
			std::cout << "Error: qbl::QDM = NULL; make sure to qbl::qdm::new first.\n";
			return true;
		}
		
		if(command_obj.getType() == wok::interp::String && wok::interp::ObjectHelper::getString(command_obj)->size() > 0)
			((QDMType*)qdm_obj.getValue())->qdm->shell(*wok::interp::ObjectHelper::getString(command_obj));
		else
			((QDMType*)qdm_obj.getValue())->qdm->shell();
	}
	else if(text[0] == "qbl::qdm::set_action_space")
	{
		// syntax is set_action_points FIELD:STRING ACTION_SPACE [ARGS]
		wok::interp::Object field_obj, space_obj, list_obj;

		interpreter->evaluate(text[1], store, field_obj, library, globalMode);
		interpreter->evaluate(text[2], store, space_obj, library, globalMode);

		if(*wok::interp::ObjectHelper::getString(field_obj) == "points" || *wok::interp::ObjectHelper::getString(field_obj) == "actions")
		{
			interpreter->evaluate(text[3], store, list_obj, library, globalMode);

			if(wok::interp::ObjectHelper::getUserObject(space_obj)->getType() != "qbl::qdm::ActionSpace"
				|| wok::interp::ObjectHelper::getUserObject(list_obj)->getType() != "qbl::qdm::ActionList")
			{
				std::cout << "Syntax error.\n";
				return true;
			}

			((ActionSpaceType*)space_obj.getValue())->A.actions = ((ActionListType*)list_obj.getValue())->actions;
		}
		else if(*wok::interp::ObjectHelper::getString(field_obj) == "degreeOfFreedom" || *wok::interp::ObjectHelper::getString(field_obj) == "movementMap")
		{
			if(wok::interp::ObjectHelper::getUserObject(space_obj)->getType() != "qbl::qdm::ActionSpace")
			{
				std::cout << "Syntax error.\n";
				return true;
			}

			((ActionSpaceType*)space_obj.getValue())->A.movementMap.clear();

			for(int i=3; i<(int)text.size(); ++i)
			{
				wok::interp::Object deg_obj;
				interpreter->evaluate(text[i], store, deg_obj, library, globalMode);
				if(deg_obj.getType() == wok::interp::Integer)
				{
					qdm::Point p;
					p.coord.push_back(*wok::interp::ObjectHelper::getInteger(deg_obj));
					((ActionSpaceType*)space_obj.getValue())->A.movementMap.push_back(p);
				}
				else if(deg_obj.getType() == wok::interp::UserType && wok::interp::ObjectHelper::getUserObject(deg_obj)->getType() == "qbl::qdm::Point")
				{
					((ActionSpaceType*)space_obj.getValue())->A.movementMap.push_back(((PointType*)deg_obj.getValue())->p);
				}
				else// if(deg_obj.getType() != wok::interp::Integer)
				{
					if(i == (int)text.size()-1 && deg_obj.getType() == wok::interp::String)
						break;
					else
					{
						std::cout << "Syntax error; degreeOfFreedom arguments must be of type Integer or qbl::qdm::Point\n";
						return true;
					}
				}

				//((ActionSpaceType*)space_obj.getValue())->A.movementMap.push_back(*wok::interp::ObjectHelper::getInteger(deg_obj));
			}
		}
		else if(*wok::interp::ObjectHelper::getString(field_obj) == "gridSize")
		{
			if(wok::interp::ObjectHelper::getUserObject(space_obj)->getType() != "qbl::qdm::ActionSpace")
			{
				std::cout << "Syntax error.\n";
				return true;
			}
			wok::interp::Object gridSize_obj;
			interpreter->evaluate(text[3], store, gridSize_obj, library, globalMode);
			((ActionSpaceType*)space_obj.getValue())->A.gridSize = *wok::interp::ObjectHelper::getInteger(gridSize_obj);
		}
		else if(*wok::interp::ObjectHelper::getString(field_obj) == "resetDelta")
		{
			if(wok::interp::ObjectHelper::getUserObject(space_obj)->getType() != "qbl::qdm::ActionSpace")
			{
				std::cout << "Syntax error.\n";
				return true;
			}
			wok::interp::Object reset_obj;
			interpreter->evaluate(text[3], store, reset_obj, library, globalMode);
			if(*wok::interp::ObjectHelper::getInteger(reset_obj) == 0)
				((ActionSpaceType*)space_obj.getValue())->A.resetDelta = false;
			else
				((ActionSpaceType*)space_obj.getValue())->A.resetDelta = true;
		}
	}
	else
		return false;
	return true;
}
}