/*
maestro-interp.h
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

/*
This program adds the functionality of usb-maestro.h to QBL (allowing one to control the functions defined in usb-maestro.h from a QBL program)
Follow this example to create other Command Libraries to add to QBL.

To actually use this, or other libraries, edit "main.cpp" (included with QBL) and add the following:
MaestroModule *m = new MaestroModule;
extras->push_back(m);

And then rebuild QBL. Note on Windows you may dynamically link to command libraries - this is demonstrated in SRS10 (not included)
*/

/*
Usage in QBL:
1) call maestro::connect 0
2) call maestro::moveServo channel:INTEGER target:DOUBLE[0,1]
3) call maestro::getAnalogInput channel:INTEGER
   returns integer
*/
#ifndef USB_MAESTRO_INTERP_MODULE_H_
#define USB_MAESTRO_INTERP_MODULE_H_

#include "usb-maestro.h"

class MaestroModule : public wok::interp::CommandLibrary
{
 public:
  virtual ~MaestroModule() {};
  virtual int requiredVersion() {return 1;};
  virtual bool newUserType(const std::string& type, const std::string& name, wok::interp::Store* store)
  {
    return false;
  }

  virtual bool evaluate(wok::Text& text, wok::interp::Store& store, wok::interp::Object& output, wok::interp::Interpreter* interpreter, std::vector <wok::interp::CommandLibrary*> &library, bool globalMode)
  {
    if(text[0] == "maestro::connect")
    {
		usbMaestro.connect(NULL);
    }
    else if(text[0] == "maestro::disconnect")
    {
		usbMaestro.disconnect();
    }
    else if(text[0] == "maestro::moveServo")
    {
		// syntax is moveServo channel:int target:double
		wok::interp::Object channel_obj, target_obj;
		interpreter->evaluate(text[1], store, channel_obj, library, globalMode);
		interpreter->evaluate(text[2], store, target_obj, library, globalMode);

		int channel = *wok::interp::ObjectHelper::getInteger(channel_obj);
		double target = *wok::interp::ObjectHelper::getDouble(target_obj);

		usbMaestro.moveServo((unsigned char)channel, (float)target);
    }
    else if(text[0] == "maestro::getAnalogInput")
    {
		// syntax is getAnalogInput channel:int
		wok::interp::Object channel_obj;
		interpreter->evaluate(text[1], store, channel_obj, library, globalMode);

		int channel = *wok::interp::ObjectHelper::getInteger(channel_obj);

		int value = usbMaestro.getAnalogInput((unsigned char)channel);

		wok::interp::ObjectHelper::newInteger(value, output);
    }
    else
      return false;
    return true;
  }

 private:
  wok::robots::Maestro usbMaestro;
};

#endif
